@@ -6,8 +6,9 @@ import ctypes
 from scipy.stats import norm as norm_dist
 from scipy.sparse import issparse, isspmatrix_csr, csr_matrix
 from sklearn.linear_model import LogisticRegression
+from sklearn.tree import DecisionTreeClassifier
 from .linreg import LinearRegression, _wrapper_double
-from ._cy_utils import _matrix_inv_symm
+from ._cy_utils import _matrix_inv_symm, _create_node_counters
 
 _unexpected_err_msg = "Unexpected error. Please open an issue in GitHub describing what you were doing."
 
@@ -224,6 +225,7 @@ def _check_active_inp(self, base_algorithm, f_grad_norm, case_one_class):
         self._force_fit = False
         self._rand_grad_norms = case_one_class
     self.case_one_class = case_one_class
+    self._force_counters = True
 
 def _check_refit_inp(refit_buffer_X, refit_buffer_r, refit_buffer):
     if (refit_buffer_X is not None) or (refit_buffer_y is not None):
@@ -349,6 +351,13 @@ def _apply_inverse_sigmoid(x):
         x[:] = np.log(x / (1.0 - x))
     return None
 
+def is_from_module(base):
+    return (isinstance(base, _BootstrappedClassifierBase) or
+            isinstance(base, _LinUCB_n_TS_single) or
+            isinstance(base, _LogisticUCB_n_TS_single) or
+            isinstance(base, _BayesianLogisticRegression) or
+            isinstance(base, _TreeUCB_n_TS_single))
+
 def _apply_softmax(x):
     x[:, :] = np.exp(x - x.max(axis=1).reshape((-1, 1)))
     x[:, :] = x / x.sum(axis=1).reshape((-1, 1))
@@ -641,19 +650,21 @@ class _OneVsRest:
                  partialfit=False, refit_buffer=0, deep_copy=False,
                  force_fit=False, force_counters=False,
                  prev_ovr=None, warm=False,
+                 force_unfit_predict=False,
                  njobs=1):
         self.n = n
         self.smooth = smooth
         self.assume_un = assume_un
         self.njobs = njobs
-        self.force_fit = force_fit
+        self.force_fit = bool(force_fit)
+        self.force_unfit_predict = bool(force_unfit_predict)
         self.thr = thr
         self.random_state = random_state
         self.refit_buffer = refit_buffer
         self.deep_copy = deep_copy
         self.partialfit = bool(partialfit)
         self.force_counters = bool(force_counters)
-        if self.force_counters or (self.thr > 0 and not self.force_fit):
+        if (self.force_counters) or (self.thr > 0 and not self.force_fit):
             ## in case it has beta prior, keeps track of the counters until no longer needed
             self.alpha = alpha
             self.beta = beta
@@ -699,12 +710,11 @@ class _OneVsRest:
                 for choice in range(self.n):
                     if isinstance(self.algos[choice], _FixedPredictor):
                         self.algos[choice] = deepcopy(base)
+                        if is_from_module(base):
+                            self.algos[choice].random_state = self.rng_arm[choice]
             else: 
                 self.algos = [deepcopy(base) for choice in range(self.n)]
-                if isinstance(base, _BootstrappedClassifierBase) \
-                   or isinstance(base, _LinUCB_n_TS_single) \
-                   or isinstance(base, _LogisticUCB_n_TS_single) \
-                   or isinstance(base, _BayesianLogisticRegression):
+                if is_from_module(base):
                     for choice in range(self.n):
                         self.algos[choice].random_state = self.rng_arm[choice]
 
@@ -723,7 +733,7 @@ class _OneVsRest:
         self.n -= 1
         if self.smooth is not None:
             self.counters = self.counters[:, np.arange(self.counters.shape[1]) != drop_ix]
-        if self.force_counters or (self.thr > 0 and not self.force_fit):
+        if (self.force_counters) or (self.thr > 0 and not self.force_fit):
             self.beta_counters = self.beta_counters[:, np.arange(self.beta_counters.shape[1]) != drop_ix]
 
     def _spawn_arm(self, fitted_classifier = None, n_w_rew = 0, n_wo_rew = 0,
@@ -734,7 +744,7 @@ class _OneVsRest:
                                 self.random_state.integers(np.iinfo(np.int32).max) + 1))
         if self.smooth is not None:
             self.counters = np.c_[self.counters, np.array([n_w_rew + n_wo_rew]).reshape((1, 1)).astype(self.counters.dtype)]
-        if self.force_counters or (self.thr > 0 and not self.force_fit):
+        if (self.force_counters) or (self.thr > 0 and not self.force_fit):
             new_beta_col = np.array([0 if (n_w_rew + n_wo_rew) < self.thr else 1, self.alpha + n_w_rew, self.beta + n_wo_rew]).reshape((3, 1)).astype(self.beta_counters.dtype)
             self.beta_counters = np.c_[self.beta_counters, new_beta_col]
         if fitted_classifier is not None:
@@ -749,7 +759,7 @@ class _OneVsRest:
                     raise ValueError("Must provide a classifier when initializing with different classifiers per arm.")
                 self.algos.append( deepcopy(self.base) )
             else:
-                if self.force_counters or (self.thr > 0 and not self.force_fit):
+                if (self.force_counters) or (self.thr > 0 and not self.force_fit):
                     self.algos.append(_BetaPredictor(self.beta_counters[:, -1][1],
                                                      self.beta_counters[:, -1][2],
                                                      self.rng_arm[-1]))
@@ -762,7 +772,7 @@ class _OneVsRest:
                 self.buffer[-1].add_obs(bufferX, buffer_y)
 
     def _update_beta_counters(self, yclass, choice):
-        if (self.beta_counters[0, choice] == 0) or self.force_counters:
+        if (self.beta_counters[0, choice] == 0) or (self.force_counters):
             n_pos = (yclass > 0.).sum()
             self.beta_counters[1, choice] += n_pos
             self.beta_counters[2, choice] += yclass.shape[0] - n_pos
@@ -791,7 +801,7 @@ class _OneVsRest:
         xclass = X[this_choice, :]
         self.algos[choice].fit(xclass, yclass)
 
-        if self.force_counters or (self.thr > 0 and not self.force_fit):
+        if (self.force_counters) or (self.thr > 0 and not self.force_fit):
             self._update_beta_counters(yclass, choice)
 
 
@@ -818,7 +828,7 @@ class _OneVsRest:
                 self.algos[choice].partial_fit(xclass, yclass, classes = [0, 1])
 
         ## update the beta counters if needed
-        if self.force_counters:
+        if (self.force_counters):
             self._update_beta_counters(yclass, choice)
 
     def _filter_arm_data(self, X, a, r, choice):
@@ -847,7 +857,8 @@ class _OneVsRest:
 
     def _decision_function_single(self, choice, X, preds, depth=2):
         ## case when using partial_fit and need beta predictions
-        if (self.partialfit or self.force_fit) and (self.thr > 0):
+        if ((self.partialfit or self.force_fit) and
+            (self.thr > 0) and (not self.force_unfit_predict)):
             if self.beta_counters[0, choice] == 0:
                 preds[:, choice] = \
                     self.rng_arm[choice].beta(self.alpha + self.beta_counters[1, choice],
@@ -915,6 +926,9 @@ class _OneVsRest:
     def get_n_neg(self, choice):
         return self.beta_counters[2, choice]
 
+    def get_counts(self):
+        return self.beta_counters[1] + self.beta_counters[2]
+
     def exploit(self, X):
         ### only usable within some policies
         pred = np.empty((X.shape[0], self.n))
@@ -1011,31 +1025,38 @@ class _LinUCB_n_TS_single:
         self.ts = ts
         self.sample_unique = bool(sample_unique)
         self.random_state = _check_random_state(random_state)
+        self.is_fitted = False
         self.model = LinearRegression(lambda_=self.lambda_,
                                       fit_intercept=self.fit_intercept,
                                       method=self.method,
-                                      use_float=self.use_float)
+                                      use_float=self.use_float,
+                                      calc_inv=True)
 
     def fit(self, X, y):
         if X.shape[0]:
             self.model.fit(X, y)
+            self.is_fitted = True
         return self
 
     def partial_fit(self, X, y, *args, **kwargs):
         if X.shape[0]:
             self.model.partial_fit(X, y)
+            self.is_fitted = True
         return self
 
     def predict(self, X, exploit=False):
         if exploit:
             return self.model.predict(X)
         elif not self.ts:
-            return self.model.predict_ucb(X, self.alpha)
+            return self.model.predict_ucb(X, self.alpha, add_unfit_noise=True,
+                                          random_state=self.random_state)
         else:
             return self.model.predict_thompson(X, self.alpha, self.sample_unique,
                                                self.random_state)
 
     def exploit(self, X):
+        if not self.is_fitted:
+            return 0.
         return self.predict(X, exploit = True)
 
 class _LogisticUCB_n_TS_single:
@@ -1049,6 +1070,7 @@ class _LogisticUCB_n_TS_single:
         self.warm_start = True
         self.sample_unique = bool(sample_unique)
         self.random_state = _check_random_state(random_state)
+        self.is_fitted = False
         self.model = LogisticRegression(C=1./lambda_, penalty="l2",
                                         fit_intercept=fit_intercept,
                                         solver='lbfgs', max_iter=15000,
@@ -1056,6 +1078,10 @@ class _LogisticUCB_n_TS_single:
         self.Sigma = np.empty((0,0), dtype=np.float64)
 
     def fit(self, X, y, *args, **kwargs):
+        if X.shape[0] == 0:
+            return self
+        elif np.unique(y).shape[0] <= 1:
+            return self
         self.model.fit(X, y)
         var = self.model.predict_proba(X)[:,1]
         var = var * (1 - var)   
@@ -1073,6 +1099,7 @@ class _LogisticUCB_n_TS_single:
             overwrite=1
         )
         _matrix_inv_symm(self.Sigma, self.lambda_)
+        self.is_fitted = True
 
     def _process_X(self, X):
         if X.dtype != ctypes.c_double:
@@ -1122,6 +1149,11 @@ class _LogisticUCB_n_TS_single:
             return pred
 
         ### UCB
+        if not self.is_fitted:
+            pred = self.conf_coef * np.sqrt(np.einsum("ij,ij->i", X, X) / self.lambda_)
+            pred[:] += self.random_state.random(size=pred.shape[0]) / 1e5
+            return pred
+
         pred = self.model.decision_function(X)
         if not exploit:
             X, Xcsr = self._process_X(X)
@@ -1130,4 +1162,85 @@ class _LogisticUCB_n_TS_single:
         return pred
 
     def exploit(self, X):
-        return 1. / (1. + np.exp(-self.model.decision_function(X)))
+        if not self.is_fitted:
+            return 0.
+        pred = self.model.decision_function(X)
+        _apply_sigmoid(pred)
+        return pred
+
+class _TreeUCB_n_TS_single:
+    def __init__(self, beta_prior=(1,1), ts=False, alpha=0.8, random_state=None,
+                 *args, **kwargs):
+        self.beta_prior = beta_prior
+        self.random_state = random_state
+        self.conf_coef = norm_dist.ppf(alpha)
+        self.ts = bool(ts)
+        self.model = DecisionTreeClassifier(*args, **kwargs)
+        self.is_fitted = False
+        self.aux_beta = (beta_prior[0], beta_prior[1])
+
+    def update_aux(self, y):
+        self.aux_beta[0] += (y >  0.).sum()
+        self.aux_beta[1] += (y <= 0.).sum()
+
+    def fit(self, X, y):
+        if X.shape[0] == 0:
+            return self
+        elif np.unique(y).shape[0] <= 1:
+            self.update_aux(y)
+            return self
+
+        seed = self.random_state.integers(np.iinfo(np.int32).max)
+        self.model.set_params(random_state = seed)
+        self.model.fit(X, y)
+        n_nodes = self.model.tree_.node_count
+        self.pos = np.zeros(n_nodes, dtype=ctypes.c_long)
+        self.neg = np.zeros(n_nodes, dtype=ctypes.c_long)
+        pred_node = self.model.apply(X).astype(ctypes.c_long)
+        _create_node_counters(self.pos, self.neg, pred_node, y.astype(ctypes.c_double))
+        self.pos = self.pos.astype(ctypes.c_double) + self.beta_prior[0]
+        self.neg = self.neg.astype(ctypes.c_double) + self.beta_prior[1]
+
+        self.is_fitted = True
+        return self
+
+    def partial_fit(self, X, y):
+        if X.shape[0] == 0:
+            return self
+        elif not self.is_fitted:
+            self.update_aux(y)
+            return self
+
+        new_pos = np.zeros(n_nodes, dtype=ctypes.c_long)
+        new_neg = np.zeros(n_nodes, dtype=ctypes.c_long)
+        pred_node = self.model.apply(X).astype(ctypes.c_long)
+        _create_node_counters(new_pos, new_neg, pred_node, y.astype(ctypes.c_double))
+        self.pos[:] += new_pos
+        self.neg[:] += new_neg
+        return self
+
+    def predict(self, X, exploit = False):
+        if not self.is_fitted:
+            n = X.shape[0]
+            if exploit:
+                return self.aux_beta[0] / (self.aux_beta[0] + self.aux_beta[1])
+            if self.ts:
+                return self.random_state.beta(self.aux_beta[0], self.aux_beta[1], size=n)
+            else:
+                mean = self.aux_beta[0] / (self.aux_beta[0] + self.aux_beta[1])
+                noise = self.random_state.random(size=n) / 1e6
+                return mean + noise
+
+        pred_node = self.model.apply(X)
+        if exploit:
+            return self.pos[pred_node] / (self.pos[pred_node] + self.neg[pred_node])
+        if self.ts:
+            return self.random_state.beta(self.pos[pred_node], self.neg[pred_node])
+        else:
+            n = self.pos[pred_node] + self.neg[pred_node]
+            mean = self.pos[pred_node] / n
+            ci = np.sqrt(mean * (1. - mean) / n)
+            return mean + self.conf_coef * ci
+
+    def exploit(self, X):
+        return self.predict(X, exploit = True)
